@@ -38,12 +38,20 @@ isnegint2(GEN a, long *pa)
 static int
 isnegint(GEN a) { return isnegint2(a, NULL); }
 static int
+isnegint_approx(GEN a, long bit)
+{
+  GEN b;
+  if (gexpo(imag_i(a)) > -bit) return 0;
+  a = real_i(a); if (gsigne(a) > 0) return 0;
+  b = ground(a); return gexpo(gsub(a, b)) < -bit;
+}
+static int
 is0(GEN a, long bit) { return gequal0(a) || gexpo(a) < -bit; }
 static int
 islong(GEN a, long *m, long prec)
 {
   *m = itos(ground(real_i(a)));
-  if (is0(gsubgs(a, *m), prec2nbits(prec))) return 1;
+  if (is0(gsubgs(a, *m), prec2nbits(prec) - 5)) return 1;
   return 0;
 }
 
@@ -68,12 +76,12 @@ airy_i(GEN x, long prec)
   if (is0(x, bit))
   {
     GEN s = sqrtnr_abs(utor(3,prec), 6), s3 = powrs(s,3), s4 = mulrr(s,s3);
-    A = invr(mulrr(s4, ggamma(sstoQ(2,3), prec)));
+    A = invr(mulrr(s4, ggamma(uutoQ(2,3), prec)));
     B = mulrr(A, s3); return mkvec2(A, B);
   }
   prec2 = prec + EXTRAPRECWORD;
   x = gprec_wensure(x, prec2);
-  z = gsqrt(gpowgs(x,3), prec2); z2 = gdivgs(gmul2n(z,1),3);
+  z = gsqrt(gpowgs(x,3), prec2); z2 = gdivgu(gmul2n(z,1),3);
   if (is_real_t(tx) && gsigne(x) > 0)
     a = b = gsqrt(x, prec2); /* expression simplifies */
   else
@@ -82,9 +90,9 @@ airy_i(GEN x, long prec)
     b = gdiv(x, a);
   }
   a = gmul(a, ibessel(sstoQ(-1,3),z2, prec));
-  b = gmul(b, ibessel(sstoQ(1,3), z2, prec));
+  b = gmul(b, ibessel(uutoQ(1,3), z2, prec));
   if (isexactzero(imag_i(x))) { a = real_i(a); b = real_i(b); }
-  A = gdivgs(gsub(a,b), 3);
+  A = gdivgu(gsub(a,b), 3);
   B = gdiv(gadd(a,b), sqrtr_abs(utor(3, prec)));
 
   bit -= gexpo(a) + 16;
@@ -201,7 +209,7 @@ hyperu_i(GEN a, GEN b, GEN x, long prec)
       {
         GEN w = gadd(gmul(gaddgs(a,k-1),u), gmul(gaddgs(qmb,1-k),v));
         u = gmul(divru(q,k),v);
-        v = gdivgs(w, k);
+        v = gdivgu(w, k);
         d = mulrr(d, c);
         e = gadd(e, gmul(d,u));
         f = gadd(f, p1 = gmul(d,v));
@@ -222,7 +230,7 @@ hyperu_i(GEN a, GEN b, GEN x, long prec)
     zf = gpow(x, gneg_i(a), l);
     for (k = n-1; k >= 0; k--)
     {
-      s = gaddsg(1, gmul(gmul(T, gdivgs(zz,k+1)), s));
+      s = gaddsg(1, gmul(gmul(T, gdivgu(zz,k+1)), s));
       if (!k) break;
       T = gsubgs(gsub(T, S), 2*k-1);
     }
@@ -349,6 +357,14 @@ vpoch(GEN a, long n)
   for (j = 1; j < n; j++) gel(v,j+1) = gmul(gel(v,j), gaddsg(j, a));
   return v;
 }
+static GEN
+RgV_vpoch(GEN N, long n)
+{
+  long i, l;
+  GEN v = cgetg_copy(N, &l);
+  for (i = 1; i < l; i++) gel(v,i) = vpoch(gel(N,i), n);
+  return v;
+}
 
 static GEN
 Npoch(GEN a, long n) { return gnorm(poch(a, n, LOWDEFAULTPREC)); }
@@ -398,7 +414,7 @@ precFtaylor(GEN N, GEN D, GEN z, long *pmi)
   {
     GEN Nj = gel(N,j);
     if (isint(Nj,&Nj) && signe(Nj) <= 0
-        && (!ma || absi_cmp(ma, Nj) < 0)) ma = Nj;
+        && (!ma || abscmpii(ma, Nj) < 0)) ma = Nj;
   }
   /* use less sharp fujiwara_bound_real(,1) ? */
   v = ground(realroots(gsub(P,Q), mkvec2(gen_0,mkoo()), pr));
@@ -446,7 +462,7 @@ Ftaylor(GEN N, GEN D, GEN z, long prec)
     GEN a = gen_1, b = gen_1;
     for (i = 1; i < lN; i++) a = gmul(a, gaddsg(j, gel(N,i)));
     for (i = 1; i < lD; i++) b = gmul(b, gaddsg(j, gel(D,i)));
-    C = gmul(C, gmul(gdiv(a, b), gdivgs(z, j+1)));
+    C = gmul(C, gmul(gdiv(a, gmulsg(j+1, b)), z));
     if (gequal0(C)) break;
     if (j > mi) tol = gequal0(S)? 0: gexpo(C) - gexpo(S);
     S = gadd(S, C); ++j;
@@ -496,14 +512,14 @@ static long
 F21ind(GEN a, GEN b, GEN c, GEN z, long bit)
 {
   GEN v = const_vec(6, mkoo());
-  long ind = 0;
+  long ind = 0, B = bit - 5;
   const long LD = LOWDEFAULTPREC;
-  if (!isnegint(cind(a,b,c, 1))) gel(v,1) = gabs(zind(z,1), LD);
+  if (!isnegint_approx(cind(a,b,c, 1),B)) gel(v,1) = gabs(zind(z,1), LD);
   gel(v,2) = gabs(zind(z,2), LD);
   gel(v,3) = gabs(z, LD);
-  if (!isnegint(cind(a,b,c, 4))) gel(v,4) = gabs(zind(z,4), LD);
-  if (!isnegint(cind(a,b,c, 5))) gel(v,5) = gabs(zind(z,5), LD);
-  if (!isnegint(cind(a,b,c, 6))) gel(v,6) = gabs(zind(z,6), LD);
+  if (!isnegint_approx(cind(a,b,c, 4),B)) gel(v,4) = gabs(zind(z,4), LD);
+  if (!isnegint_approx(cind(a,b,c, 5),B)) gel(v,5) = gabs(zind(z,5), LD);
+  if (!isnegint_approx(cind(a,b,c, 6),B)) gel(v,6) = gabs(zind(z,6), LD);
   ind = vecindexmin(v); /* |znew| <= 1; close to 1 ? */
   return (gexpo(gsubgs(gel(v,ind),1)) > -maxss(bit / 4, 32))? -ind: ind;
 }
@@ -551,9 +567,9 @@ F21finitetaylor(long m, GEN b, GEN c, GEN z, long prec)
 {
   pari_sp av;
   GEN C, S;
-  long j, ct, pradd, mi, tol, bitmin, mb;
-  if (isnegint2(b, &mb) && mb < m) { b = stoi(-m); m = mb; }
-  pradd = precFtaylor(mkvec2(stoi(-m), b), mkvec(c), z, &mi);
+  long j, ct, pradd, mi, bitmin, mb;
+  if (isnegint2(b, &mb) && mb < m) { b = utoineg(m); m = mb; }
+  pradd = precFtaylor(mkvec2(utoineg(m), b), mkvec(c), z, &mi);
   if (pradd > 0)
   {
     prec += pradd;
@@ -562,16 +578,15 @@ F21finitetaylor(long m, GEN b, GEN c, GEN z, long prec)
     z = gprec_wensure(z, prec);
   }
   bitmin = -(prec2nbits(prec) + 10);
-  C = real_1(prec); S = C; ct = 0; tol = 0;
+  C = real_1(prec); S = C; ct = 0;
   av = avma;
   for(j = 0; j < m; ++j)
   {
-    C = gmul(C, gdiv(gmulsg(-(m-j), gaddsg(j, b)), gaddsg(j, c)));
-    C = gmul(C, gdivgs(z, j+1));
-    if (j > mi) tol = gequal0(S) ? 0: gexpo(C) - gexpo(S);
+    C = gmul(C, gdiv(gmulsg(j-m, gaddsg(j, b)), gmulsg(j+1, gaddsg(j, c))));
+    C = gmul(C, z);
+    if (j > mi && !gequal0(S))
+    { if (gexpo(C) - gexpo(S) > bitmin) ct = 0; else if (++ct == 3) break; }
     S = gadd(S, C);
-    if (j > mi)
-    { if (tol > bitmin) ct = 0; else if (++ct == 3) break; }
     if (gc_needed(av, 1)) gerepileall(av, 2, &S, &C);
   }
   return S;
@@ -703,6 +718,11 @@ F21_i(GEN a, GEN b, GEN c, GEN z, long prec)
   }
   /* Here a, b, c, c-a, c-b are not nonpositive integers */
   ind = F21ind(a, b, c, z, bitprec);
+  prec2 = prec + EXTRAPREC64;
+  a = gprec_wensure(a,prec2);
+  b = gprec_wensure(b,prec2);
+  c = gprec_wensure(c,prec2);
+  z = gprec_wensure(z,prec2);
   if (ind < 0) return gmul(ggamma(c, prec), F21taylorind(a,b,c, z, ind, prec));
   if (gsigne(real_i(b)) <= 0)
   {
@@ -720,11 +740,6 @@ F21_i(GEN a, GEN b, GEN c, GEN z, long prec)
   }
   if (gcmp(real_i(a), real_i(b)) < 0 && gsigne(real_i(a)) > 0) swap(a,b);
   /* Here real(b) > 0 and either real(a) <= 0 or real(a) > real(b) */
-  prec2 = prec + EXTRAPREC64;
-  a = gprec_wensure(a,prec2);
-  b = gprec_wensure(b,prec2);
-  c = gprec_wensure(c,prec2);
-  z = gprec_wensure(z,prec2);
   if (gcmp(real_i(c), real_i(b)) <= 0)
   {
     long N = 1 + itos(gfloor(gsub(real_i(b),real_i(c)))); /* >= 1 */
@@ -759,13 +774,17 @@ F21(GEN a, GEN b, GEN c, GEN z, long prec)
 
 static GEN
 F21taylor(GEN a, GEN b, GEN c, GEN z, long prec)
-{ return gdiv(Ftaylor(mkvec2(a,b), mkvec(c), z, prec), ggamma(c, prec)); }
+{
+  pari_sp av = avma;
+  GEN r = gdiv(Ftaylor(mkvec2(a,b), mkvec(c), z, prec), ggamma(c, prec));
+  return gerepileupto(av, r);
+}
 
 static GEN
 F21taylorlim(GEN N, long m, GEN z, GEN Z, long ind, long prec)
 {
   pari_sp av;
-  GEN C, P, S, tmp;
+  GEN C, P, S, a, b;
   long j, ct, pradd, mi, fl, bitmin, tol, si = (ind == 5 || ind == 6)? -1: 1;
   pradd = precFtaylor(N, mkvec(stoi(m + 1)), z, &mi);
   if (pradd)
@@ -775,30 +794,27 @@ F21taylorlim(GEN N, long m, GEN z, GEN Z, long ind, long prec)
     z = gprec_wensure(z, prec);
     Z = gprec_wensure(Z, prec);
   }
+  av = avma; a = gel(N,1); b = gel(N,2);
   bitmin = -(prec2nbits(prec) + 10);
-  P = glog(Z, prec);
-  if (ind == 4 || ind == 5) P = gneg(P);
+  P = glog(Z, prec); if (ind == 4 || ind == 5) P = gneg(P);
   P = gadd(P, gsub(gpsi(stoi(m+1), prec), mpeuler(prec)));
-  tmp = gel(N, 2); if (si == -1) tmp = gsubsg(1, tmp);
-  P = gsub(P, gadd(gpsi(gel(N, 1), prec), gpsi(tmp, prec)));
-  C = real_1(prec); ct = 0; j = 0; tol = 0;
-  S = P; fl = 1;
-  av = avma;
-  for(;;)
+  P = gsub(P, gadd(gpsi(a, prec), gpsi(si == -1? gsubsg(1, b): b, prec)));
+  C = real_1(prec); ct = 0; tol = 0; S = P;
+  for(j = 0, fl = 1;;)
   {
-    GEN v1 = gaddsg(j, gel(N,1)), v2 = gaddsg(j, gel(N,2));
-    long jB = (j+1) * (j+1+m);
-    C = gdivgs(gmul(z, gmul(C, v1)), jB);
+    GEN v1 = gaddgs(a, j), v2 = gaddgs(b, j);
+    long J = (j+1) * (j+1+m);
+    C = gmul(C, gdivgs(gmul(z, v1), J));
     if (gequal0(v2)) fl = 0; else C = gmul(C, v2);
     if (j > mi) tol = gequal0(S) ? 0 : gexpo(C) - gexpo(S);
     if (fl)
     {
-      P = gadd(P, gsub(sstoQ(2*j+2+m, jB), gadd(ginv(v1), ginv(v2))));
+      P = gadd(P, gsub(uutoQ(2*j+2+m, J), gadd(ginv(v1), ginv(v2))));
       S = gadd(S, gmul(C, P));
     }
-    else S = (si == 1)? gadd(S, C): gsub(S, C);
-    if (++j > mi)
-    { if (tol > bitmin) ct = 0; else if (++ct == 3) break; }
+    else
+      S = (si == 1)? gadd(S, C): gsub(S, C);
+    if (++j > mi) { if (tol > bitmin) ct = 0; else if (++ct == 3) break; }
     if (gc_needed(av, 1)) gerepileall(av, 3, &S, &C, &P);
   }
   return gdiv(S, mpfact(m));
@@ -1023,7 +1039,7 @@ F21taylorind(GEN a, GEN b, GEN c, GEN z, long ind, long prec)
   return gprec_wtrunc(res, prec);
 }
 
-static void
+static long
 hypersimplify(GEN *pn, GEN *pd)
 {
   GEN n = *pn, d = *pd;
@@ -1036,10 +1052,10 @@ hypersimplify(GEN *pn, GEN *pd)
       if (gequal(t, gel(n, k)))
       {
         *pd = vecsplice(d, j);
-        *pn = vecsplice(n, k); return hypersimplify(pd, pn);
+        *pn = vecsplice(n, k); return hypersimplify(pd, pn) + 1;
       }
   }
-  return;
+  return 0;
 }
 
 static GEN
@@ -1064,10 +1080,7 @@ f_pochall(void *E, GEN n)
 }
 static GEN
 f_pochall_alt(void *E, GEN n)
-{
-  GEN z = f_pochall(E,n);
-  return mpodd(n)? gneg(z): z;
-}
+{ GEN z = f_pochall(E,n); return mpodd(n)? gneg(z): z; }
 /* z = \pm1 */
 static GEN
 sumz(GEN N, GEN D, long z, long prec)
@@ -1086,19 +1099,15 @@ hypergeom_arg(GEN x)
   if (!x) return cgetg(1,t_VEC);
   return (typ(x) == t_VEC)? x: mkvec(x);
 }
+
 static GEN
 hypergeom_i(GEN N, GEN D, GEN z, long prec)
 {
-  long nN, nD, j;
+  long nN, nD;
   if (!is_scalar_t(typ(z))) pari_err_TYPE("hypergeom",z);
   if (gequal0(z)) return gen_1;
-  N = hypergeom_arg(N);
-  D = hypergeom_arg(D);
-  hypersimplify(&N, &D);
   nN = lg(N) - 1;
   nD = lg(D) - 1;
-  for (j = 1; j <= nD; ++j)
-    if (isnegint(gel(D,j))) pari_err_TYPE("hypergeom", D);
   if (nD >= (nN? nN: 2)) return Ftaylor(N, D, z, prec);
   if (nD == nN - 1 && nN >= 3)
   {
@@ -1128,6 +1137,77 @@ hypergeom_i(GEN N, GEN D, GEN z, long prec)
   pari_err_IMPL("this hypergeometric function");
   return NULL; /*LCOV_EXCL_LINE*/
 }
+
+/* [v[i]+k, i=1..#v]; no need to optimize for k = 0 */
+static GEN
+RgV_z_add(GEN v, long k)
+{
+  long i, l;
+  GEN w = cgetg_copy(v, &l);
+  for (i = 1; i < l; i++) gel(w,i) = gaddgs(gel(v,i), k);
+  return w;
+}
+static GEN
+vpoch_mul(GEN v, long k)
+{
+  long i, l = lg(v);
+  GEN p;
+  if (l == 1) return gen_1;
+  p = gmael(v, 1, k);
+  for (i = 2; i < l; i++) p = gmul(p, gmael(v, i, k));
+  return p;
+}
+
+static GEN
+serhypergeom(GEN N, GEN D, GEN y, long prec)
+{
+  pari_sp av;
+  GEN S, pn, vN, vD, y0 = NULL;
+  long v, l, i;
+  if (!signe(y)) return gadd(gen_1, y);
+  v = valp(y);
+  if (v < 0) pari_err_DOMAIN("hypergeom","valuation", "<", gen_0, y);
+  l = lg(y);
+  if (v) S = gen_1;
+  else
+  {
+    y0 = gel(y, 2);
+    y = serchop0(y);
+    l = 3 + (l - 3) / valp(y);
+    S = hypergeom(N, D, y0, prec);
+  }
+  vN = RgV_vpoch(N, l-1);
+  vD = RgV_vpoch(D, l-1); av = avma;
+  pn = y;
+  for (i = 1; i < l; i++)
+  {
+    GEN H = gdiv(vpoch_mul(vN, i), vpoch_mul(vD, i));
+    if (y0) H = gmul(H, hypergeom_i(RgV_z_add(N,i), RgV_z_add(D,i), y0,prec));
+    S = gadd(S, gmul(pn, H)); if (i + 1 < l) pn = gdivgu(gmul(pn, y), i + 1);
+    if (gc_needed(av,1))
+    {
+      if (DEBUGMEM>1) pari_warn(warnmem,"hypergeom, i = %ld / %ld", i,l-1);
+      gerepileall(av, 2, &S, &pn);
+    }
+  }
+  return S;
+}
+
 GEN
-hypergeom(GEN N, GEN D, GEN z, long prec)
-{ pari_sp av = avma; return gerepilecopy(av, hypergeom_i(N,D,z,prec)); }
+hypergeom(GEN N, GEN D, GEN y, long prec)
+{
+  pari_sp av = avma;
+  GEN z;
+  long j, n;
+  N = hypergeom_arg(N);
+  D = hypergeom_arg(D);
+  n = hypersimplify(&N, &D);
+  for (j = 1; j < lg(D); j++)
+    if (isnegint(gel(D,j)))
+      pari_err_DOMAIN("hypergeom", stack_sprintf("b[%ld]", j + n),
+                      "<=", gen_0, gel(D,j));
+  if (is_scalar_t(typ(y)))
+    return gerepilecopy(av, hypergeom_i(N, D, y, prec));
+  if (!(z = toser_i(y))) pari_err_TYPE("hypergeom", y);
+  return gerepileupto(av, serhypergeom(N, D, z, prec));
+}
